#!/usr/bin/php
<?php

// Copyright (C) 2023 Cygnus <cygnus2 at post.cz>
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or (at
// your option) any later version.
// 
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA

// BASH colors
define ("CFAIL","\033[1;31;40m !\033[0m ");
define ("COK","\033[1;32;40m *\033[0m ");
define ("CWARN","\033[1;33;40m ?\033[0m ");
define ("CTRY","\033[1;34;40m -\033[0m ");

define ("CRED","\033[1;31;40m");
define ("CGRN","\033[1;32;40m");
define ("CYELL","\033[1;33;40m");
define ("CBLUE","\033[1;34;40m");
define ("CMAGN","\033[1;35;40m");
define ("CCYAN","\033[1;36;40m");
define ("CWHITE","\033[1;37;40m");

define ("CDEF","\033[0m");

require ("./text2hpgl_config.php");

$outbasename = "";	// will contain filename without extension

// conversion table from ISO8859-2 to UTF-8 (characters 160..256)
$set_iso88592 = [
		' ', 'Ą', '˘', 'Ł', '¤', 'Ľ', 'Ś', '§', '¨', 'Š', 'Ş', 'Ť', 'Ź', ' ', 'Ž', 'Ż',
		'°', 'ą', '˛', 'ł', '´', 'ľ', 'ś', 'ˇ', '¸', 'š', 'ş', 'ť', 'ź', '˝', 'ž', 'ż',
		'Ŕ', 'Á', 'Â', 'Ă', 'Ä', 'Ĺ', 'Ć', 'Ç', 'Č', 'É', 'Ę', 'Ë', 'Ě', 'Í', 'Î', 'Ď',
		'Đ', 'Ń', 'Ň', 'Ó', 'Ô', 'Ő', 'Ö', '×', 'Ř', 'Ů', 'Ú', 'Ű', 'Ü', 'Ý', 'Ţ', 'ß',
		'ŕ', 'á', 'â', 'ă', 'ä', 'ĺ', 'ć', 'ç', 'č', 'é', 'ę', 'ë', 'ě', 'í', 'î', 'ď',
		'đ', 'ń', 'ň', 'ó', 'ô', 'ő', 'ö', '÷', 'ř', 'ů', 'ú', 'ű', 'ü', 'ý', 'ţ', '˙'
		];


//-------------------------------------------------------------------------------------------------

function draw_glyph (& $outfile, $x, $y, $c, $scale = CFG_SCALE)
{
	global $font;
	
	$pairs = $font[$c][0];
	$pen = 0;
	
	// move pen on first coordinate with pen up
	fwrite($outfile, "PU".($x + $scale*$font[$c][2][0]).",".($y + $scale*$font[$c][2][1]).";");
	if (CFG_DEBUG) echo ($x + $scale*$font[$c][2][0]).",".($y + $scale*$font[$c][2][1])."; ";
	$pen = 1;
	
	for ($i = 0; $i < $pairs; $i++)
	{
		if ((($font[$c][2+$i][0]) == -1) && (($font[$c][2+$i][1]) == -1))
		{
			$pen = 0;
		}
		else
		{
			if ($pen == 0)
				fwrite($outfile, "PU");
			else
				fwrite($outfile, "PD");
				
			fwrite($outfile, ($x + $scale*$font[$c][2+$i][0]).",".($y + $scale*$font[$c][2+$i][1]).";");
			
			if ($pen == 0)
			{
				$pen = 1;
				fwrite($outfile, "PD".($x + $scale*$font[$c][2+$i][0]).",".($y + $scale*$font[$c][2+$i][1]).";");
			}
		}
	}
	fwrite($outfile,"\n");
}

function print_line (& $outfile, & $line,& $x,& $y, $align = 'left', $indent = 0)
{
	global $font;
	
	$pen = 0;	// 0 = up, 1 = down
	$i = 0;		// character index
	
	$l = strlen($line);
	
	if ($l > 0)
	{
		$spaces = 0;
		$i = 0;
		$w = 0;
		
		// first line of paragraph
		if ($indent == 1 && PARAGRAPH_INDENT)
		{
			$w += CFG_SCALE * CFG_CHARW * PARAGRAPH_INDENT_WIDTH;
		}

		// compute number of spaces in text line
		while ($w < (CFG_PAPERW - CFG_MARGIN_LEFT))
		{
			$w += CFG_SCALE * $font[ord($line[$i])][1];
			if (ord($line[$i]) == 32) $spaces++;
			
			$i++;
			if ($i > ($l - 1)) break;
		}
		if (CFG_DEBUG) echo CRED."DIF=".(CFG_PAPERW - (CFG_MARGIN_LEFT + $w)).CDEF;
		if (CFG_DEBUG) echo " I=".$indent." SP=".$spaces." W=".$w." ";
		
		if ($spaces > 0)
		{
			$sm = (CFG_PAPERW - CFG_MARGIN_LEFT - $w) % $spaces;
			$sf = floor((CFG_PAPERW - CFG_MARGIN_LEFT - $w) / $spaces);
			
			$sp = array();
			for ($i = 0; $i < $spaces; $i++)
			{
				$sp[$i] = $sf;
				if ($sm > 0) { $sp[$i]++; $sm--; }
			}
		}
	
		$i = 0;
		$si = 0;
		
		while ($l > $i)
		{
			if (CFG_DEBUG)
			{
				echo CYELL.$line[$i].CDEF;
				echo $font[ord($line[$i])][1]." ";		// width
				echo CGRN.$font[ord($line[$i])][0].CDEF." ";	// pairs
			}
			
			if (($x + CFG_SCALE * $font[ord($line[$i])][1]) > CFG_PAPERW)
				break;
			
			// draw glyph
			if (ord($line[$i]) != 32)
				draw_glyph ($outfile, $x, $y, ord($line[$i]));

			// move to the right by width of printed character
			$x += CFG_SCALE * $font[ord($line[$i])][1];
			
			// make space wider if aligning text to the both margins
			if ((ord($line[$i]) == 32) && ($align == 'justify') && ($spaces > 0))
			{
				$x += $sp[$si];
				
				if (CFG_DEBUG) echo CRED."SP=".$sp[$si].CDEF;
				$si++;
			}
				
			$i++;
			
			// check if line width was exceeded
			if ($x > CFG_PAPERW) echo CRED."long!".CDEF;
		}
		
		$line = substr($line, $i);
		$line = trim($line);
	}
		
	// substr('abcdef', 1);     // bcdef
	// UTF-8 https://www.php.net/manual/en/function.mb-substr.php
}

function hpgl_width ($txt)
{
	global $font;
	$maxchar = 32 + count($font);	// different fonts can have different limits
	
	$w = 0;
	
	if (strlen($txt) > 0)
	{
		$i = 0;
		
		while ($i < strlen($txt))
		{
			// for now it works only with ASCII characters without diacritics, because font doesn't contain anything else
			if ((ord($txt[$i]) > 31) && (ord($txt[$i]) <= $maxchar))
			{
				$w += CFG_SCALE*$font[ord($txt[$i])][1];	// add width of character
			}
			$i++;
		}
	}
	
	return ($w);
}

function print_aligned_line ($outfile, $line, $column, $txt, $align = "left", $scale = CFG_SCALE)
{
	// align = left / center
	
	$w = hpgl_width ($txt);
	switch ($align)
	{
		case "center":
			// ignore column
			$x = CFG_MARGIN_LEFT + round((CFG_PAPERW - CFG_MARGIN_LEFT - $w)/2);
			break;
		default:
			$x = CFG_MARGIN_LEFT + $scale * CFG_CHARW * $column; 
			break;
	}
	
	$y = CFG_PAPERH - $scale*($line*CFG_LINEH);
	
	print_line ($outfile, $txt , $x, $y, 'left');
	
	return ($w);
}

function generate_fontpage ($outfile)
{
	global $font;
	
	// start page
	fwrite($outfile, "IN;SP0;PU0,0;\n");
	
	// this is function which will print whole font, even ISO8859-2 extension if font have it
	// 96 characters 32..127	12 x 8?
	// 96 characters 160..255
	
	$maxchar = 32 + count($font);	// last character in loaded font
	$centering = CFG_MARGIN_LEFT + round((CFG_PAPERW - CFG_MARGIN_LEFT - (12*1.8*CFG_CHARW)*CFG_SCALE*2) / 2);
	
	// characters 32..127
	$character = 32;	// start with space in standard ASCII
	for ($row = 0; $row < 8; $row++)
		for ($col = 0; $col < 12; $col++)
		{
			if ($character < $maxchar)
			{
				$x = round($centering + ($col*1.8*CFG_CHARW)*CFG_SCALE*2);
				$y = round(CFG_PAPERH - CFG_LINEH*CFG_SCALE*2*$row);
				draw_glyph ($outfile, $x, $y, $character, CFG_SCALE*2);
			}
			$character++;
		}
	
	// characters 160..255
	$character = 160;	// first character in ISO8859-2 extension
	for ($row = 0; $row < 8; $row++)
		for ($col = 0; $col < 12; $col++)
		{
			if ($character < $maxchar)
			{
				$x = round($centering + ($col*1.8*CFG_CHARW)*CFG_SCALE*2);
				$y = round(CFG_PAPERH - CFG_LINEH*CFG_SCALE*2*($row+10));
				draw_glyph ($outfile, $x, $y, $character, CFG_SCALE*2);
			}
			$character++;
		}
	
	// finish page
	fwrite($outfile, "\nPU0,0;SP0;IN; \n \n \n \n \n \n");
}

function generate_testpage ($outfile)
{
	// start page
	fwrite($outfile, "IN;SP0;PU0,0;\n");
	
	// this is function for verification of specific parameter for selected plotter with selected font
	
	// draw rectangle around printable area
	fwrite($outfile, "PD0,".CFG_PAPERH.";");
	fwrite($outfile, "PD".CFG_PAPERW.",".CFG_PAPERH.";");
	fwrite($outfile, "PD".CFG_PAPERW.",0;");
	fwrite($outfile, "PD0,0;PU;");
	
	// draw lines for bottom margin and left margin
	fwrite($outfile, "PU0,".CFG_MARGIN_BOTTOM.";");
	fwrite($outfile, "PD".CFG_PAPERW.",".CFG_MARGIN_BOTTOM.";PU;");
	
	fwrite($outfile, "PU".CFG_MARGIN_LEFT.",0;");
	fwrite($outfile, "PD".CFG_MARGIN_LEFT.",".CFG_PAPERH.";PU;");
	
	// print centered title and two lines for underscore
	print_aligned_line ($outfile, 3, 0, "TXT to HPGL converter - TEST PAGE", "center");
	$w = print_aligned_line ($outfile, 4, 0, "program written by ZXCygnus 2023", "center");
	
	$x = CFG_MARGIN_LEFT + round(((CFG_PAPERW - CFG_MARGIN_LEFT) - $w)/2);
	$y = round(CFG_PAPERH - CFG_SCALE*(4*CFG_LINEH+CFG_LINEH/3));
	
	fwrite($outfile, "PU".$x.",".$y.";");
	fwrite($outfile, "PD".($x+$w).",".$y.";");
	
	$y = $y - round(CFG_SCALE*4);
	
	fwrite($outfile, "PU".$x.",".$y.";");
	fwrite($outfile, "PD".($x+$w).",".$y.";");
	
	// print configuration parameters
	print_aligned_line ($outfile, 8, 5, "CFG_CHARW", "left");
	print_aligned_line ($outfile, 8, 18, strval(CFG_CHARW), "left");
	print_aligned_line ($outfile, 9, 5, "CFG_LINEH", "left");
	print_aligned_line ($outfile, 9, 18, strval(CFG_LINEH), "left");
	print_aligned_line ($outfile, 10, 5, "CFG_SCALE", "left");
	print_aligned_line ($outfile, 10, 18, strval(CFG_SCALE), "left");
	print_aligned_line ($outfile, 12, 5, "CFG_MARGIN_LEFT", "left");
	print_aligned_line ($outfile, 12, 28, strval(CFG_MARGIN_LEFT), "left");
	print_aligned_line ($outfile, 13, 5, "CFG_MARGIN_BOTTOM  ", "left");
	print_aligned_line ($outfile, 13, 28, strval(CFG_MARGIN_BOTTOM), "left");
	print_aligned_line ($outfile, 15, 5, "CFG_PAPERW x CFG_PAPERH  ".strval(CFG_PAPERW)." x ".strval(CFG_PAPERH), "left");
	
	// print ASCII table aligned at monospace raster
	$centering = CFG_MARGIN_LEFT + round((CFG_PAPERW - CFG_MARGIN_LEFT - (16*1.8*CFG_CHARW)*CFG_SCALE*2) / 2);
	for ($r = 0; $r < 4; $r++)
	{
		for ($i = 0; $i < 16; $i++)
		{
			$x = round($centering + ($i*1.8*CFG_CHARW)*CFG_SCALE*2);
			$y = round(CFG_PAPERH - 16*(CFG_LINEH+3*$r)*CFG_SCALE*2);
			draw_glyph ($outfile, $x, $y, 32+$i+16*$r, CFG_SCALE*2);
		}
	}
	
	// finish page
	fwrite($outfile, "\nPU0,0;SP0;IN; \n \n \n \n \n \n");
}

// ------------------------------------------------------------------------------------------------

function nextpage (& $outfile)
{
	global $outbasename;
	
	fclose ($outfile);
	
	$pageindex = 1;
	
	if (file_exists ($outbasename."_".$pageindex.".hpgl"))
	{
		while (file_exists ($outbasename."_".$pageindex.".hpgl"))
			$pageindex++;
	}
	
	$outfile = fopen ($outbasename."_".$pageindex.".hpgl", "w");
}

function txt2html (& $infile, & $outfile)
{
	global $font;
	
	// default position should be [0,0] - pen at left bottom corner - go to the top left corner
	fwrite($outfile, "IN;SP0;PU0,".CFG_PAPERH.";\n");
	
	$x = CFG_MARGIN_LEFT;	// starting pen position - top left page corner
	$y = CFG_PAPERH;
	
	$line = "";		// reset line
	$linew = 0;
	$linen = 1;		// line number
	$word = "";		// reset word 
	$wordw = 0;
	$newparagraph = 0;
	
	$maxchar = 32 + count($font);	// different fonts can have different limits
	
	if (PARAGRAPH_INDENT)
	{
		// paragraph indentation - add width of 3 monospace characters
		$x = CFG_MARGIN_LEFT + CFG_SCALE * CFG_CHARW * PARAGRAPH_INDENT_WIDTH;
		$linew = CFG_SCALE * CFG_CHARW * PARAGRAPH_INDENT_WIDTH;
		$newparagraph = 1;
	}
	
	while (1) 
	{
		$char = fread($infile, 1);	// read one character from input file

		// check for end of file
		if ($char === false || feof($infile))
		{
			// finish remaining word & line
			if (strlen($word) > 0)
			{
				if (strlen ($line) == 0)
				{
					$line .= $word;
					$linew .= $wordw;
				}
				else
				{
					$line .= ' '.$word;
					$linew = $linew + CFG_SCALE*$font[32][1] + $wordw;
				}
			}
			if (strlen($line) > 0) print_line ($outfile, $line, $x, $y, 'left');
			
			// finish HPGL page
			fwrite($outfile, "PU0,0;SP0;IN; \n \n \n \n \n \n");
			break;
		}
		
		// is it ASCII character? (for now without diacritics)
		if ((ord($char) > 32) && (ord($char) <= $maxchar))
		{
			$word .= $char;					// add character to the word
			$wordw += CFG_SCALE * $font[ord($char)][1];	// add width of character
		}

		// end of word
		if (($char == ' ') || (ord($char) == 10))
		{
			if (CFG_DEBUG && (ord($char) == 10)) echo CCYAN."\n".$linen." ".$y." ".CDEF;
			
			// compute width of word before adding it to the line
			$w = $wordw;
			if (strlen($line) > 0) $w += CFG_SCALE*$font[32][1];
			
			// will word fit on line?
			if (($linew + $w) < (CFG_PAPERW - CFG_MARGIN_LEFT))
			{
				// if word is not first on the line then add space before word
				if (strlen($line) > 0)
				{
					$word = ' '.$word;
					$wordw += CFG_SCALE*$font[32][1];
				}
				
				if (CFG_DEBUG) echo "+ ".$word;

				$line .= $word;
				$linew += $w;
				
				$word = "";
				$wordw = 0;
				
				// end of line -> print line (probably shorter)
				if (ord($char) == 10)
				{
					print_line ($outfile, $line, $x, $y, 'left');
					$line = "";
					$linew = 0;
					$linen++;
					
					$x = CFG_MARGIN_LEFT;
					$y = $y - CFG_SCALE*CFG_LINEH;
					
					if (PARAGRAPH_INDENT)
					{
						// paragraph indentation - add width of 3 monospace characters
						$x = CFG_MARGIN_LEFT + CFG_SCALE * CFG_CHARW * PARAGRAPH_INDENT_WIDTH;
						$linew = CFG_SCALE * CFG_CHARW * PARAGRAPH_INDENT_WIDTH;
					}
					
					if ($y < CFG_MARGIN_BOTTOM)
					{
						if (CFG_DEBUG) echo CMAGN."---- full page ----\n".CDEF;
						
						// finish HPGL page
						fwrite($outfile, "PU0,0;SP0;IN; \n \n \n \n \n \n");
						
						// create  new page
						nextpage ($outfile);
						
						fwrite($outfile, "IN;SP0;PU0,".CFG_PAPERH.";\n");
						
						$x = CFG_MARGIN_LEFT;	// starting pen position - top left page corner
						$y = CFG_PAPERH;
						
						if (PARAGRAPH_INDENT)
						{
							// paragraph indentation - add width of 3 monospace characters
							$x = CFG_MARGIN_LEFT + CFG_SCALE * CFG_CHARW * PARAGRAPH_INDENT_WIDTH;
							$linew = CFG_SCALE * CFG_CHARW * PARAGRAPH_INDENT_WIDTH;
						}
					}
					
					$newparagraph = 1;
				}
			}
			else
			{
				// $linew + $w >= CFG_PAPERW
				// word doesn't fit on line:
				// - line is already full, cannot add another word - print first then add word
				// - line is empty but word is too long - print first then add word
				if (CFG_DEBUG) echo CCYAN."\n".$linen." ".$y." ".CBLUE."full ln\n".CDEF;
				
				// will print what will fit, and then will remove only printed part from line
				print_line ($outfile, $line, $x, $y, 'justify', $newparagraph);
				$line = "";
				$linew = 0;
				$linen++;

				$x = CFG_MARGIN_LEFT;
				$y = $y - CFG_SCALE * CFG_LINEH;	// one line down by character height * scale
				
				$newparagraph = 0;
			
				if ($y < CFG_MARGIN_BOTTOM)
				{
					if (CFG_DEBUG) echo CMAGN."---- full page ----\n".CDEF;
					
					// finish HPGL page
					fwrite($outfile, "PU0,0;SP0;IN; \n \n \n \n \n \n");
					
					// create  new page
					nextpage ($outfile);
					
					fwrite($outfile, "IN;SP0;PU0,".CFG_PAPERH.";\n");
					
					$x = CFG_MARGIN_LEFT;	// starting pen position - top left page corner
					$y = CFG_PAPERH;
				}
				
				// line is printed and shorter of it's original content, add word
				// if word is not first on the line then add space before word
				if ((strlen($line) > 0) && (strlen($line, -1) != ' '))
				{
					$word = ' '.$word;
					$wordw += CFG_SCALE*$font[32][1];
				}
				
				if (CFG_DEBUG) echo "+ ".$word;

				$line .= $word;
				$linew += $w;
				
				$word = "";
				$wordw = 0;

				if (CFG_DEBUG) echo CYELL.$linen." ".$wordw." ".$word.CDEF." \n";
			}
		}
	}
	if (CFG_DEBUG) echo "\n";
}

if (isset ($argv[1]))
{
	$option_count=0;
	foreach ($argv as $key => $option)
	{
		// ignore $option_count[0], because it is filename of script itself
		if (($option_count > 0) && ($option[0] != '-'))
		{
			if (file_exists ($option))
			{
				// make sure existing file will not be rewritten
				$bname = basename($option,".txt");
				$fileindex = 0;
				
				if (file_exists ($bname."_".$fileindex.".hpgl"))
				{
					while (file_exists ($bname."_".$fileindex.".hpgl")) $fileindex++;
				}
				
				$outname = $bname."_".$fileindex.".hpgl";
				$outbasename = $bname;
				
				$in = fopen ($option, "r");
				$out = fopen ($outname, "w");
				
				if ($in && $out)
				{
					if (CFG_DEBUG) echo CWHITE."Opening filename: ".CYELL.$option."\n".CDEF;
					if (CFG_DEBUG) echo CWHITE."Creating output file: ".CYELL.$outname."\n".CDEF;
					
					txt2html ($in, $out);
				}
				else
				{
					if (!$in) echo CRED."Opening input file ".$option." failed.\n".CDEF;
					if (!$out) echo CRED."Opening output file ".$outname." failed.\n".CDEF;
				}
				if (isset($in)) fclose ($in);
				if (isset($out)) fclose ($out);
			}
			else
			{
				echo "File ".$option." not found.\n";
			}
		}
		
		if (($option_count > 0) && ($option[0] == '-'))
		{
			if ($option == "-h") {
				echo "This program will convert plain text in to very basic HPGL usable with many plotters especially mine. \nAll details how to use this program are in readme.txt which should be included.\n";
				echo "\n";
				echo "Parameters:\n";
				echo "-h\twill return to the console this simple help and basic information about this program\n";
				echo "-t\twill generate testpage using values from servected config\n";
				echo "-T\twill generate fontpage with whole font including ISO8859-2 extension if exists\n";
			}
			
			if ($option == "-t")
			{
				// make sure existing file will not be rewritten
				$testpagename = "testpage";
				$fileindex = 0;
				
				if (file_exists ($testpagename."_".$fileindex.".hpgl"))
				{
					while (file_exists ($testpagename."_".$fileindex.".hpgl")) $fileindex++;
				}
				
				$outname = $testpagename."_".$fileindex.".hpgl";
				
				echo "creating test page into file ".$outname."\n";
				
				$out = fopen ($outname, "w");
				generate_testpage ($out);
				fclose ($out);
				echo "\n";
			}
			
			if ($option == "-T")
			{
				// make sure existing file will not be rewritten
				$testpagename = "fontpage";
				$fileindex = 0;
				
				if (file_exists ($testpagename."_".$fileindex.".hpgl"))
				{
					while (file_exists ($testpagename."_".$fileindex.".hpgl")) $fileindex++;
				}
				
				$outname = $testpagename."_".$fileindex.".hpgl";
				
				if (count($font) < 96+128)
				{
					echo CRED."Selected font doesn't contain all characters.\n".CDEF;
					echo CRED."Fontpage will be limited.\n\n".CDEF;
				}
				echo "creating font page into file ".$outname."\n";
				
				$out = fopen ($outname, "w");
				generate_fontpage ($out);
				fclose ($out);
				echo "\n";
			}
			
// 			if ($option == "-w")
// 			{
// 				echo "\n".$option[$key+1]."\n";
// 				if (is_numeric($option[$key+1]))
// 				{
// 					$paper_width = $option[$key+1];
// 					if ($paper_width < 100) $paper_width = CFG_PAPERW;
// 				}
// 				else
// 				{
// 					echo "Wrong width.\n";
// 					return (-1);
// 				}
// 			}
		}
		$option_count++;
	}

}
else
	echo CFAIL."You must specify at least one filename for input file.\n".CDEF;

return(0);

?>
